@@ -360,15 +360,18 @@ static int matches (struct itemplate *t, unsigned char *data, int asize,
 	if (c >= 064 && c <= 066) {
 	    ins->oprs[c-064].offset = *data++;
 	    ins->oprs[c-064].offset |= (*data++ << 8);
-	    if (asize == 32) {
+	    if (osize == 32) {
 		ins->oprs[c-064].offset |= (((long) *data++) << 16);
 		ins->oprs[c-064].offset |= (((long) *data++) << 24);
 		ins->oprs[c-064].segment |= SEG_32BIT;
 	    } else
 		ins->oprs[c-064].segment &= ~SEG_32BIT;
 	    ins->oprs[c-064].segment |= SEG_RELATIVE;
-	    if (segsize != asize)
-		ins->oprs[c-064].addr_size = asize;
+	    if (segsize != osize) {
+	        ins->oprs[c-064].type =
+		    (ins->oprs[c-064].type & NON_SIZE)
+		    | ((osize == 16) ? BITS16 : BITS32);
+	    }
 	}
 	if (c >= 070 && c <= 072) {
 	    ins->oprs[c-070].offset = *data++;
@@ -377,13 +380,23 @@ static int matches (struct itemplate *t, unsigned char *data, int asize,
 	    ins->oprs[c-070].offset |= (((long) *data++) << 24);
 	    ins->oprs[c-070].segment |= SEG_32BIT | SEG_RELATIVE;
 	}
-	if (c >= 0100 && c <= 0177) {
+	if (c >= 0100 && c < 0130) {
 	    int modrm = *data++;
 	    ins->oprs[c & 07].basereg = (modrm >> 3) & 07;
 	    ins->oprs[c & 07].segment |= SEG_RMREG;
 	    data = do_ea (data, modrm, asize, segsize,
 			  &ins->oprs[(c >> 3) & 07]);
 	}
+	if (c >= 0130 && c <= 0132) {
+	    ins->oprs[c-0130].offset = *data++;
+	    ins->oprs[c-0130].offset |= (*data++ << 8);
+	}
+	if (c >= 0140 && c <= 0142) {
+	    ins->oprs[c-0140].offset = *data++;
+	    ins->oprs[c-0140].offset |= (*data++ << 8);
+	    ins->oprs[c-0140].offset |= (((long) *data++) << 16);
+	    ins->oprs[c-0140].offset |= (((long) *data++) << 24);
+	}
 	if (c >= 0200 && c <= 0277) {
 	    int modrm = *data++;
 	    if (((modrm >> 3) & 07) != (c & 07))
