@@ -11,9 +11,7 @@
 #ifndef NASM_NASM_H
 #define NASM_NASM_H
 
-#define NASM_MAJOR_VER 0
-#define NASM_MINOR_VER 98
-#define NASM_VER "0.98"
+#include "version.h"		       /* generated NASM version macros */
 
 #ifndef NULL
 #define NULL 0
@@ -33,6 +31,16 @@
 #define FILENAME_MAX 256
 #endif
 
+#ifndef PREFIX_MAX
+#define PREFIX_MAX 10
+#endif
+
+#ifndef POSTFIX_MAX
+#define POSTFIX_MAX 10
+#endif
+
+
+
 /*
  * Name pollution problems: <time.h> on Digital UNIX pulls in some
  * strange hardware header file which sees fit to define R_SP. We
@@ -57,33 +65,38 @@ struct ofmt;
 /*
  * An error reporting function should look like this.
  */
-typedef void (*efunc) (int severity, char *fmt, ...);
+typedef void (*efunc) (int severity, const char *fmt, ...);
 
 /*
  * These are the error severity codes which get passed as the first
  * argument to an efunc.
  */
 
-#define ERR_WARNING 0		       /* warn only: no further action */
-#define ERR_NONFATAL 1		       /* terminate assembly after phase */
-#define ERR_FATAL 2		       /* instantly fatal: exit with error */
-#define ERR_PANIC 3		       /* internal error: panic instantly
+#define ERR_DEBUG  	0x00000008	/* put out debugging message */
+#define ERR_WARNING	0x00000000	/* warn only: no further action */
+#define ERR_NONFATAL	0x00000001	/* terminate assembly after phase */
+#define ERR_FATAL	0x00000002	/* instantly fatal: exit with error */
+#define ERR_PANIC	0x00000003	/* internal error: panic instantly
 					* and dump core for reference */
-#define ERR_MASK 0x0F		       /* mask off the above codes */
-#define ERR_NOFILE 0x10		       /* don't give source file name/line */
-#define ERR_USAGE 0x20		       /* print a usage message */
-#define ERR_PASS1 0x80		       /* only print this error on pass one */
+#define ERR_MASK	0x0000000F	/* mask off the above codes */
+#define ERR_NOFILE	0x00000010	/* don't give source file name/line */
+#define ERR_USAGE	0x00000020	/* print a usage message */
+#define ERR_PASS1	0x00000040	/* only print this error on pass one */
 
 /*
  * These codes define specific types of suppressible warning.
  */
-#define ERR_WARN_MNP  0x0100	       /* macro-num-parameters warning */
-#define ERR_WARN_OL   0x0200	       /* orphan label (no colon, and
-					* alone on line) */
-#define ERR_WARN_NOV  0x0300	       /* numeric overflow */
-#define ERR_WARN_MASK 0xFF00	       /* the mask for this feature */
+
+#define ERR_WARN_MASK	0x0000FF00	/* the mask for this feature */
 #define ERR_WARN_SHR  8		       /* how far to shift right */
-#define ERR_WARN_MAX  3		       /* the highest numbered one */
+
+#define ERR_WARN_MNP	0x00000100	/* macro-num-parameters warning */
+#define ERR_WARN_MSR	0x00000200	/* macro self-reference */
+#define ERR_WARN_OL	0x00000300	/* orphan label (no colon, and
+					* alone on line) */
+#define ERR_WARN_NOV	0x00000400	/* numeric overflow */
+#define ERR_WARN_GNUELF	0x00000500      /* using GNU ELF extensions */
+#define ERR_WARN_MAX	5		/* the highest numbered one */
 
 /*
  * -----------------------
@@ -250,8 +263,8 @@ struct eval_hints {
  * defined before use", whereas if `critical' is 2, the error will
  * be "symbol undefined".
  *
- * If `critical' has bit 4 set (in addition to its main value: 0x11
- * and 0x12 correspond to 1 and 2) then an extended expression
+ * If `critical' has bit 8 set (in addition to its main value: 0x101
+ * and 0x102 correspond to 1 and 2) then an extended expression
  * syntax is recognised, in which relational operators such as =, <
  * and >= are accepted, as well as low-precedence logical operators
  * &&, ^^ and ||.
@@ -259,6 +272,7 @@ struct eval_hints {
  * If `hints' is non-NULL, it gets filled in with some hints as to
  * the base register in complex effective addresses.
  */
+#define CRITICAL 0x100
 typedef expr *(*evalfunc) (scanner sc, void *scprivate, struct tokenval *tv,
 			   int *fwref, int critical, efunc error,
 			   struct eval_hints *hints);
@@ -297,7 +311,7 @@ typedef struct {
     /*
      * Called at the end of a pass.
      */
-    void (*cleanup) (void);
+    void (*cleanup) (int);
 } Preproc;
 
 /*
@@ -368,6 +382,7 @@ enum {
 
 #define TO        0x00000100L          /* reverse effect in FADD, FSUB &c */
 #define COLON     0x00000200L	       /* operand is followed by a colon */
+#define STRICT    0x00000400L	       /* do not optimize this operand */
 
 /* type of operand: memory reference, register, etc. */
 #define MEMORY    0x00204000L
@@ -411,7 +426,9 @@ enum {
 /* special type of immediate operand */
 #define ONENESS   0x00800000L          /* so UNITY == IMMEDIATE | ONENESS */
 #define UNITY     0x00802000L	       /* for shift/rotate instructions */
-
+#define BYTENESS  0x40000000L          /* so SBYTE == IMMEDIATE | BYTENESS */
+#define SBYTE 	  0x40002000L	       /* for op r16/32,immediate instrs. */
+		
 /*
  * Next, the codes returned from the parser, for registers and
  * instructions.
@@ -526,19 +543,19 @@ struct ofmt {
      * This is a short (one-liner) description of the type of
      * output generated by the driver.
      */
-    char *fullname;
+    const char *fullname;
 
     /*
      * This is a single keyword used to select the driver.
      */
-    char *shortname;
+    const char *shortname;
 
     /*
      * this is reserved for out module specific help.
      * It is set to NULL in all the out modules but is not implemented
      * in the main program
      */
-    char *helpstring;
+    const char *helpstring;
 
     /*
      * this is a pointer to the first element of the debug information
@@ -742,12 +759,12 @@ struct dfmt {
      * This is a short (one-liner) description of the type of
      * output generated by the driver.
      */
-    char *fullname;
+    const char *fullname;
 
     /*
      * This is a single keyword used to select the driver.
      */
-    char *shortname;
+    const char *shortname;
 
 
     /*
@@ -839,4 +856,17 @@ struct dfmt {
 
 #define elements(x)     ( sizeof(x) / sizeof(*(x)) )
 
+extern int tasm_compatible_mode;
+
+/*
+ * This declaration passes the "pass" number to all other modules
+ * "pass0" assumes the values: 0, 0, ..., 0, 1, 2
+ * where 0 = optimizing pass
+ *       1 = pass 1
+ *       2 = pass 2
+ */
+
+extern int pass0;	/* this is globally known */
+extern int optimizing;
+
 #endif
