@@ -19,6 +19,10 @@
 #include "parser.h"
 #include "float.h"
 
+extern int  in_abs_seg;		/* ABSOLUTE segment flag */
+extern long abs_seg;		/* ABSOLUTE segment */
+extern long abs_offset;		/* ABSOLUTE segment offset */
+
 static long reg_flags[] = {	       /* sizes and special flags */
     0, REG8, REG_AL, REG_AX, REG8, REG8, REG16, REG16, REG8, REG_CL,
     REG_CREG, REG_CREG, REG_CREG, REG_CR4, REG_CS, REG_CX, REG8,
@@ -34,7 +38,7 @@ static long reg_flags[] = {	       /* sizes and special flags */
 
 enum {				       /* special tokens */
     S_BYTE, S_DWORD, S_FAR, S_LONG, S_NEAR, S_NOSPLIT, S_QWORD,
-    S_SHORT, S_TO, S_TWORD, S_WORD
+    S_SHORT, S_STRICT, S_TO, S_TWORD, S_WORD
 };
 
 static int is_comma_next (void);
@@ -99,7 +103,7 @@ insn *parse_line (int pass, char *buffer, insn *result,
 	     * Generally fix things. I think this is right as it is, but
 	     * am still not certain.
 	     */
-	    ldef (result->label, location->segment,
+	    ldef (result->label, in_abs_seg?abs_seg:location->segment,
 		  location->offset, NULL, TRUE, FALSE, outfmt, errfunc);
 	}
     }
@@ -122,7 +126,7 @@ insn *parse_line (int pass, char *buffer, insn *result,
 	    expr *value;
 
 	    i = stdscan(NULL, &tokval);
-	    value = evaluate (stdscan, NULL, &tokval, NULL, pass, error, NULL);
+	    value = evaluate (stdscan, NULL, &tokval, NULL, pass0, error, NULL);
 	    i = tokval.t_type;
 	    if (!value) {	       /* but, error in evaluator */
 		result->opcode = -1;   /* unrecoverable parse error: */
@@ -189,9 +193,10 @@ insn *parse_line (int pass, char *buffer, insn *result,
 	result->opcode == I_RESD ||
 	result->opcode == I_RESQ ||
 	result->opcode == I_REST ||
-	result->opcode == I_EQU)
+	result->opcode == I_EQU ||
+        result->opcode == I_INCBIN)  /* fbk */
     {
-	critical = pass;
+	critical = pass0;
     }
     else
 	critical = (pass==2 ? 2 : 0);
@@ -403,6 +408,9 @@ insn *parse_line (int pass, char *buffer, insn *result,
 	      case S_TO:
 		result->oprs[operand].type |= TO;
 		break;
+	      case S_STRICT:
+		result->oprs[operand].type |= STRICT;
+		break;
 	      case S_FAR:
 		result->oprs[operand].type |= FAR;
 		break;
@@ -421,27 +429,63 @@ insn *parse_line (int pass, char *buffer, insn *result,
 	if (i == '[' || i == '&') {    /* memory reference */
 	    mref = TRUE;
 	    bracket = (i == '[');
-	    i = stdscan(NULL, &tokval);	    
+	    i = stdscan(NULL, &tokval);
 	    if (i == TOKEN_SPECIAL) {  /* check for address size override */
-		switch ((int)tokval.t_integer) {
-		  case S_NOSPLIT:
-		    result->oprs[operand].eaflags |= EAF_TIMESTWO;
-		    break;
-		  case S_BYTE:
-		    result->oprs[operand].eaflags |= EAF_BYTEOFFS;
-		    break;
-		  case S_WORD:
-		    result->oprs[operand].addr_size = 16;
-		    result->oprs[operand].eaflags |= EAF_WORDOFFS;
-		    break;
-		  case S_DWORD:
-		  case S_LONG:
-		    result->oprs[operand].addr_size = 32;
-		    result->oprs[operand].eaflags |= EAF_WORDOFFS;
-		    break;
-		  default:
-		    error (ERR_NONFATAL, "invalid size specification in"
-			   " effective address");
+		if (tasm_compatible_mode) {
+		  switch ((int)tokval.t_integer) {
+		    /* For TASM compatibility a size override inside the
+		     * brackets changes the size of the operand, not the
+		     * address type of the operand as it does in standard
+		     * NASM syntax. Hence:
+		     *
+		     *	mov	eax,[DWORD val]
+		     *
+		     * is valid syntax in TASM compatibility mode. Note that
+		     * you lose the ability to override the default address
+		     * type for the instruction, but we never use anything
+		     * but 32-bit flat model addressing in our code.
+		     */
+		    case S_BYTE:
+		      result->oprs[operand].type |= BITS8;
+		      break;
+		    case S_WORD:
+		      result->oprs[operand].type |= BITS16;
+		      break;
+		    case S_DWORD:
+		    case S_LONG:
+		      result->oprs[operand].type |= BITS32;
+		      break;
+		    case S_QWORD:
+		      result->oprs[operand].type |= BITS64;
+		      break;
+		    case S_TWORD:
+		      result->oprs[operand].type |= BITS80;
+		      break;
+		    default:
+		      error (ERR_NONFATAL, "invalid operand size specification");
+		  }
+		} else {
+		  /* Standard NASM compatible syntax */
+		  switch ((int)tokval.t_integer) {
+		    case S_NOSPLIT:
+		      result->oprs[operand].eaflags |= EAF_TIMESTWO;
+		      break;
+		    case S_BYTE:
+		      result->oprs[operand].eaflags |= EAF_BYTEOFFS;
+		      break;
+		    case S_WORD:
+		      result->oprs[operand].addr_size = 16;
+		      result->oprs[operand].eaflags |= EAF_WORDOFFS;
+		      break;
+		    case S_DWORD:
+		    case S_LONG:
+		      result->oprs[operand].addr_size = 32;
+		      result->oprs[operand].eaflags |= EAF_WORDOFFS;
+		      break;
+		    default:
+		      error (ERR_NONFATAL, "invalid size specification in"
+			     " effective address");
+		  }
 		}
 		i = stdscan(NULL, &tokval);
 	    }
@@ -549,7 +593,7 @@ insn *parse_line (int pass, char *buffer, insn *result,
 
 		else if (e->value != 1)          /* If both want to be index */
 		{
-		    error(ERR_NONFATAL, "invalid effective address");
+		    error(ERR_NONFATAL, "beroset-p-592-invalid effective address");
 		    result->opcode = -1;
 		    return result;
 		} 
@@ -560,7 +604,7 @@ insn *parse_line (int pass, char *buffer, insn *result,
 	    if (e->type != 0) {	       /* is there an offset? */
 		if (e->type <= EXPR_REG_END)  /* in fact, is there an error? */
 		{
-		    error (ERR_NONFATAL, "invalid effective address");
+		    error (ERR_NONFATAL, "beroset-p-603-invalid effective address");
 		    result->opcode = -1;
 		    return result;
 		} 
@@ -587,14 +631,14 @@ insn *parse_line (int pass, char *buffer, insn *result,
 			 * Look for a segment base type.
 			 */
 			if (e->type && e->type < EXPR_SEGBASE) {
-			    error (ERR_NONFATAL, "invalid effective address");
+			    error (ERR_NONFATAL, "beroset-p-630-invalid effective address");
 			    result->opcode = -1;
 			    return result;
 			}
 			while (e->type && e->value == 0)
 			    e++;
 			if (e->type && e->value != 1) {
-			    error (ERR_NONFATAL, "invalid effective address");
+			    error (ERR_NONFATAL, "beroset-p-637-invalid effective address");
 			    result->opcode = -1;
 			    return result;
 			}
@@ -607,7 +651,7 @@ insn *parse_line (int pass, char *buffer, insn *result,
 			while (e->type && e->value == 0)
 			    e++;
 			if (e->type) {
-			    error (ERR_NONFATAL, "invalid effective address");
+			    error (ERR_NONFATAL, "beroset-p-650-invalid effective address");
 			    result->opcode = -1;
 			    return result;
 			}
@@ -620,7 +664,7 @@ insn *parse_line (int pass, char *buffer, insn *result,
 	    }
 
 	    if (e->type != 0) {    /* there'd better be nothing left! */
-		error (ERR_NONFATAL, "invalid effective address");
+		error (ERR_NONFATAL, "beroset-p-663-invalid effective address");
 		result->opcode = -1;
 		return result;
 	    }
@@ -647,13 +691,19 @@ insn *parse_line (int pass, char *buffer, insn *result,
 		result->oprs[operand].offset = reloc_value(value);
 		result->oprs[operand].segment = reloc_seg(value);
 		result->oprs[operand].wrt = reloc_wrt(value);
-		if (is_simple(value) && reloc_value(value)==1)
-		    result->oprs[operand].type |= UNITY;
+		if (is_simple(value)) {
+		    if (reloc_value(value)==1)
+			result->oprs[operand].type |= UNITY;
+		    if (optimizing>=0 &&
+			!(result->oprs[operand].type & STRICT)) {
+  		        if (reloc_value(value) >= -128 &&
+		                 reloc_value(value) <= 127)
+		            result->oprs[operand].type |= SBYTE;
+		    }
+		}
 	    } 
 	    else	       /* it's a register */
 	    {
-		int i;
-
 		if (value->type>=EXPR_SIMPLE || value->value!=1) {
 		    error (ERR_NONFATAL, "invalid operand type");
 		    result->opcode = -1;
